#!/usr/bin/python

from configparser import ConfigParser
import os

current = None
location = None 

def generate():
    """
    Convenience function to create a default SADFace configuration file if one is needed
    """
    config = ConfigParser()
    
    if not os.path.exists('new.cfg'):
        config['analyst'] = {'name': 'A User', 'email': 'user@email.address'}
        config['file'] = {'name': 'data', 'dir': 'deploy/out/'}
        config['db'] = {'name':'argdb', 'type':'couchdb', 'ip':'127.0.0.1', 'port':'5984', 'protocol':'http'}
    
    config.write(open('new.cfg', 'w'))


def init(location=None):
    """
    Function to initialise the SADFace configuration
    """
    if(location is not None):
        set_location(location)
        load()
    else:
        reset()


def load():
    """
    Load a configuration file from the stored config_location
    """
    if(location is not None):
        try:
            global current
            current = ConfigParser()
            current.read(location)
        except:
            print("Could not read configs from " + location)
            exit(1) 
    else:
        raise Exception("Tried to load config file but location is set to None")
        exit(1)

def reset():
    """
    Return config parameters to initial settings
    """
    global current
    global location
    current = None
    location = None

def set_location(new_location):
    """
    Enable the location of custom configuration files to be supplied
    """
    global location
    location = new_location





import configparser
import json
import os
import unittest

import sadface as sf

class TestVerification(unittest.TestCase):
    def setUp(self):
        sf.reset()

    def tearDown(self):
        sf.reset()

    def test_verify(self):
        """
        TESTS: sadface.validation.verify()
        """
        pass

        minimal_good = {'edges':[],'metadata':{'core':{'analyst_email':'user@email.address','analyst_name':'A User','created':'2020-06-25T18:05:36','edited':'2020-06-25T18:05:36','id':'235bfdee-2406-4c40-97cc-6f4c2539c69e','version':'0.2'}},'nodes':[],'resources':[]}

        result, problems = sf.validation.verify(minimal_good)
        self.assertEqual(True, result)
        self.assertEqual([], problems)
        
        """
        # No specified configuration file. location is None
        with self.assertRaises(Exception) as context:
            sf.config.load()
            
        self.assertTrue("Tried to load config file but location is set to None" in str(context.exception))

        # With bad configuration file
        sf.config.set_location("src/sadface/aml.py")
        with self.assertRaises(SystemExit):
            sf.config.load()
        
        # With good configuration file
        sf.config.set_location("etc/test.cfg")
        sf.config.load()
        sf.init()
        out = sf.sd
        out['metadata']['core']['created'] = ""
        out['metadata']['core']['edited'] = ""
        out['metadata']['core']['id'] = ""
        expected = json.loads('{ "edges": [], "metadata": { "core": { "analyst_email": "you-killed-my-father@prepare-to-die.com", "analyst_name": "Inigo Montoya", "created": "", "edited": "", "id": "", "version": "0.2" } }, "nodes": [], "resources": []}')
        self.assertEqual(out, expected)
        """


if __name__ == "__main__":
    
    unittest.main()



import xml.etree.ElementTree as et
from . import sadface

def init(infile):
    sadface.sd = sadface.initialise()
    tree = et.parse(infile)
    root = tree.getroot()
    process_root(root)

def process_root(root):
    c = root.getchildren()
    for child in c:
        if child.tag == "TEXT":
            process_text(child)
            pass
        elif child.tag == "SCHEMESET":
            pass
        elif child.tag == "EDATA":
            pass
        elif child.tag == "AU":
            process_au(child)
        else:
            print("XML malformed")

def process_au(node):
    ch = node.getchildren()
    for c in ch:
        if c.tag == "PROP":
            pass
        elif c.tag == "REFUTATION":
            process_refutation(c, node)
        elif c.tag == "CA":
            process_ca(c, node)
        elif c.tag == "LA":
            process_la(c, node)
        else:
            print("XML malformed")

def process_ca(node, parent):
    """
    The DTD allows only a single premise supporting a given conclusion
    within a single AU node.
    """
    conc = get_node_text(parent)[1]
    prem = []

    ch = node.getchildren()
    for c in ch:
        if c.tag == "AU":
            prem.append(get_node_text(c)[1])
            process_au(c)

    sadface.add_support(con_text=conc, prem_text=prem)

def process_la(node, parent):
    conc = get_node_text(parent)[1]
    prem = []

    ch = node.getchildren()
    for c in ch:
        if c.tag == "AU":
            prem.append(get_node_text(c)[1])
            process_au(c)

    sadface.add_support(con_text=conc, prem_text=prem)

def process_refutation(node, parent):
    """
    The DTD allows only a single refutation relationship between a
    pair of AU nodes.
    """
    conc = get_node_text(parent)[1]
    prem = []

    ch = node.getchildren()
    for c in ch:
        if c.tag == "AU":
            prem.append(get_node_text(c)[1])
            process_au(c)

    sadface.add_conflict(arg_text=conc, conflict_text=prem[0])

def get_node_text(node):
    for child in node:
        if child.tag == "PROP":
            pid = child.attrib['identifier']
            for elem in child.iter(tag="PROPTEXT"):
                return (pid, elem.text)

def process_text(node):
    sadface.add_resource(node.text)




import datetime
import json
import os
import unittest

from uuid import UUID

import sadface as sf

class TestCore(unittest.TestCase):
    def setUp(self):
        sf.reset()

    def tearDown(self):
        sf.reset()

    def test_build_argument(self):
        """
        TESTS: sadface.build_argument(con_text=None, prem_text=None, con_id=None, prem_id=None)
        """
        sf.initialise()
        con1 = "You should treasure every moment"
        prem1 = ["if you are going to die then you should treasure every moment", "You are going to die"]
        arg1 = sf.build_argument(con_text=con1, prem_text=prem1, con_id=None, prem_id=None)

        self.assertEqual(con1, arg1.get("conclusion").get("text"))
        
        prem1_atom = sf.get_atom(arg1.get("premises")[0])
        self.assertEqual(prem1[0], prem1_atom.get("text"))

        prem2_atom = sf.get_atom(arg1.get("premises")[1])
        self.assertEqual(prem1[1], prem2_atom.get("text"))

    def test_build_disagreement(self):
        """
        TESTS: build_disagreement(arg_text=None, arg_id=None, disagreement_text=None, disagreement_id=None)
        """
        sf.initialise()
        a_text = "roses are red"
        c_text = "roses are white"
        disagreement = sf.build_disagreement(arg_text=a_text, disagreement_text=c_text)

        self.assertEqual(a_text, disagreement.get("argument_1").get("text"))
        self.assertEqual(c_text, disagreement.get("argument_2").get("text"))

        self.assertEqual("disagree", disagreement.get("conflict").get("name"))


    def test_add_support(self):
        """
        TESTS: add_support(con_text=None, prem_text=None, con_id=None, prem_id=None)
        """
        sf.initialise()
        con1 = "You should treasure every moment"
        prem1 = ["if you are going to die then you should treasure every moment", "You are going to die"]
        arg1 = sf.add_support(con_text=con1, prem_text=prem1, con_id=None, prem_id=None)

        self.assertEqual(con1, arg1.get("conclusion").get("text"))
        
        prem1_atom = sf.get_atom(arg1.get("premises")[0])
        self.assertEqual(prem1[0], prem1_atom.get("text"))

        prem2_atom = sf.get_atom(arg1.get("premises")[1])
        self.assertEqual(prem1[1], prem2_atom.get("text"))


    def test_add_resource(self):
        """
        TESTS: sadface.add_resource(content)
        """
        sf.initialise()
        new_resource = sf.add_resource("DAKA DAKA")
        new_resource_content = new_resource.get("content")
        new_resource_type = new_resource.get("type")
        
        self.assertTrue(new_resource.get("id"))
        out = new_resource.get("id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

        self.assertTrue(new_resource.get("metadata"))
        self.assertTrue(type(new_resource.get("metadata")) is dict)

        expected = {"core"}
        self.assertEqual(set(expected), set(new_resource.get("metadata")))


        self.assertEqual(new_resource_content, "DAKA DAKA")
        self.assertEqual(new_resource_type, "text")

    def test_add_resource_metadata(self):
        """
        TESTS: sadface.add_resource_metadata()
        """
        sf.initialise()

         # Add a resource
        resource_text = "test resource"
        resource = sf.add_resource(resource_text)
        resource_id = resource.get("id")

        # Check resource metadata is empty
        resource = sf.get_resource(resource_id)
        self.assertEqual(resource_id, resource.get("id"))
        meta = resource.get("metadata").get("core")
        self.assertEqual(0, len(meta))

        # add metadata to core namespace
        sf.add_resource_metadata(resource_id, "core", "KEY1", "VALUE1")
        resource = sf.get_resource(resource_id)
        meta = resource.get("metadata").get("core")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))

        # add metadata to a new namespace
        sf.add_resource_metadata(resource_id, "META1", "KEY1", "VALUE1")
        resource = sf.get_resource(resource_id)
        meta = resource.get("metadata").get("META1")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))

    def test_add_source(self):
        """
        TESTS: sadface.add_source()
        """
        sf.initialise()

        # Add an atom
        text = "DAKA DAKA"
        atom = sf.add_atom(text)
        atom_id =  sf.contains_atom(text)
        
         # Add a resource
        resource_text = "test resource"
        resource = sf.add_resource(resource_text)
        resource_id = resource.get("id")

        # Add source to the atom, referencing the resource
        offset = 5
        length = len(resource_text)
        sf.add_source(atom_id, resource_id, resource_text, offset, length)

        # Now retrieve the source and test it
        s = sf.get_source(atom_id, resource_id)
        self.assertEqual(s.get("resource_id"), resource_id)
        self.assertEqual(s.get("text"), resource_text)
        self.assertEqual(s.get("offset"), offset)
        self.assertEqual(s.get("length"), length)

    def test_add_atom(self):
        """
        Tests: sadface.add_atom()
        """
        sf.initialise()

        # Check we have no atoms in the default document
        num_atoms = len(sf.list_atoms())
        self.assertEqual(num_atoms, 0)

        # Add an atom then check how many atoms we have
        atom_text = "test atom"
        atom = sf.add_atom(atom_text)
        atom_id = atom.get("id")
        num_atoms = len(sf.list_atoms())
        self.assertEqual(num_atoms, 1)

        # Retrieve the new atom and check that it
        # contains the expected text
        atom = sf.get_atom(atom_id)
        self.assertEqual(atom.get("text"), atom_text)

    def test_add_edge(self):
        """
        TESTS: sadface.add_edge()
        """
        sf.initialise()
        source = sf.add_atom("source")
        target = sf.add_atom("target")
        edge = sf.add_edge(source.get("id"), target.get("id"))

        retrieved_edge = sf.get_edge(edge.get("id"))
        self.assertEqual(retrieved_edge.get("source_id"), source.get("id"))
        self.assertEqual(retrieved_edge.get("target_id"), target.get("id"))
        self.assertEqual(retrieved_edge.get("id"), edge.get("id"))

    def test_add_atom_metadata(self):
        """
        TESTS: sadface.add_atom_metadata()
        """
        sf.initialise()

         # Add an atom
        atom_text = "test atom"
        atom = sf.add_atom(atom_text)
        atom_id = atom.get("id")

        # Check atom metadata is empty
        atom = sf.get_atom(atom_id)
        self.assertEqual(atom_id, atom.get("id"))
        meta = atom.get("metadata").get("core")
        self.assertEqual(0, len(meta))

        # add core metadata
        sf.add_atom_metadata(atom_id, "core", "KEY1", "VALUE1")
        atom = sf.get_atom(atom_id)
        meta = atom.get("metadata").get("core")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))

        # add metadata to new namespace
        sf.add_atom_metadata(atom_id, "META1", "KEY1", "VALUE1")
        atom = sf.get_atom(atom_id)
        meta = atom.get("metadata").get("META1")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))


    def test_add_global_metadata(self):
        """
        TESTS: sadface.add_global_metadata()
        """
        sf.initialise()

        # add core metadata
        sf.add_global_metadata("core", "KEY1", "VALUE1")
        sd = sf.get_document()
        meta = sd.get("metadata").get("core")
        self.assertEqual("VALUE1", meta.get("KEY1"))

        # add metadata to new namespace
        sf.add_global_metadata("META1", "KEY1", "VALUE1")
        sd = sf.get_document()
        meta = sd.get("metadata").get("META1")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))


    def test_add_notes(self):
        """
        Tests: sadface.add_notes()
        """
        sf.initialise()
        self.assertEqual(sf.get_notes(), None)

        text = "DAKA DAKA"
        sf.add_notes(text)
        self.assertEqual(sf.get_notes(), text)

        text2 = "MORE DAKA"
        sf.add_notes(text2)
        self.assertEqual(sf.get_notes(), text2)


    def test_add_inference(self):
        """
        TESTS: sadface.add_inference() with default values

        Add an inference, then retrieve it, and ensure that the 
        retrieved inference matches that which was added
        """
        sf.initialise()
        inference = sf.add_inference("test inference")
        inference_id = inference.get("id")
        result = sf.get_inference(inference_id)
        result_id = result.get("id")
        self.assertEqual(result_id, inference_id)

    def test_add_inference_metadata(self):
        """
        TESTS: sadface.add_inference_metadata()
        """
        sf.initialise()

         # Add an inference
        inference_text = "test inference"
        inference = sf.add_inference(inference_text)
        inference_id = inference.get("id")

        # Check inference metadata is empty
        inference = sf.get_inference(inference_id)
        self.assertEqual(inference_id, inference.get("id"))
        meta = inference.get("metadata").get("core")
        self.assertEqual(0, len(meta))

        # add core metadata
        sf.add_inference_metadata(inference_id, "core", "KEY1", "VALUE1")
        inference = sf.get_inference(inference_id)
        meta = inference.get("metadata").get("core")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))

        # add metadata to new namespace
        sf.add_inference_metadata(inference_id, "META1", "KEY1", "VALUE1")
        inference = sf.get_inference(inference_id)
        meta = inference.get("metadata").get("META1")
        self.assertNotEqual(0, len(meta))
        self.assertEqual("VALUE1", meta.get("KEY1"))

    def test_append_notes(self):
        """
        TESTS: sadface.append_notes()
        """
        sf.initialise()
        text = "DAKA DAKA"
        sf.append_notes(text)
        self.assertEqual(sf.get_notes(), text)
        
        text2 = "MORE DAKA"
        sf.append_notes(text2)
        self.assertEqual(sf.get_notes(), text+text2)

    def test_clear_notes(self):
        """
        Tests: sadface.clear_notes()
        """
        sf.initialise()
        self.assertEqual(sf.get_notes(), None)

        text = "DAKA DAKA"
        sf.add_notes(text)
        self.assertEqual(sf.get_notes(), text)

        sf.clear_notes()
        self.assertEqual(sf.get_notes(), None)

    def test_contains_atom(self):
        """
        TESTS: sadface.contains_atom(atom_text)
        """
        sf.initialise()

        # Test Retrieving atom that doesn't exist
        self.assertEqual(None, sf.contains_atom("DAKA DAKA"))

        # Add new atom then test retrieving by text
        text = "DAKA DAKA"
        atom = sf.add_atom(text)
        atom_id =  sf.contains_atom(text)
        retrieved_atom = sf.get_atom(atom_id)
        self.assertEqual(atom, retrieved_atom)

    def test_delete_atom(self):
        """
        TESTS: sadface.delete_atom(atom_id)
        """
        sf.initialise()

        # Remove non-existing atom
        with self.assertRaises(ValueError):
            sf.delete_atom("TESTID")

        # Remove known existing atom
        atom = sf.add_atom("TESTATOM")
        retrieved = sf.get_atom(atom.get("id"))
        self.assertEqual(atom, retrieved)
        sf.delete_atom(atom.get("id"))
        retrieved = sf.get_atom(atom.get("id"))
        self.assertEqual(None, retrieved)

    def test_delete_edge(self):
        """
        TESTS: sadface.delete_edge(edge_id)
        """
        sf.initialise()

        # Remove non-existing edge
        with self.assertRaises(ValueError):
            sf.delete_edge("TESTID")
        
        # Remove known existing atom
        source_atom = sf.add_atom("SOURCE")
        target_atom = sf.add_atom("TARGET")
        edge = sf.add_edge(source_atom.get("id"), target_atom.get("id"))
        retrieved = sf.get_edge(edge.get("id"))
        self.assertEqual(edge, retrieved)

        sf.delete_edge(edge.get("id"))
        retrieved = sf.get_edge(edge.get("id"))
        self.assertEqual(None, retrieved)
        
    def test_delete_source(self):
        """
        TESTS: sadface.delete_source(atom_id, resource_id)
        """
        sf.initialise()

        # setup
        a = sf.add_atom("ATOM")
        r = sf.add_resource("hello")
        s = sf.add_source(a.get("id"), r.get("id"), "TEXT", 10, 4)

        # verify existing source
        retrieved = sf.get_source(a.get("id"), r.get("id"))
        self.assertEqual(s.get("id"), retrieved.get("id"))
        self.assertEqual(s, retrieved)

        # Delete source
        sf.delete_source(a.get("id"), r.get("id"))
        retrieved = sf.get_source(a.get("id"), r.get("id"))
        self.assertEqual(None, retrieved)


    def test_delete_resource(self):
        """
        TESTS: sadface.delete_resource(esource_id)
        """
        sf.initialise()

        # Remove non-existing resource
        with self.assertRaises(ValueError):
            sf.delete_resource("FAKE_ID")

        # Remove existing resource
        r = sf.add_resource("NEW RESOURCE")
        retrieved = sf.get_resource(r.get("id"))
        self.assertEqual(r, retrieved)

        sf.delete_resource(r.get("id"))

        retrieved = sf.get_resource(r.get("id"))
        self.assertEqual(None, retrieved)


    def test_delete_inference(self):
        """
        """
        sf.initialise()

        # Remove non-existing inference
        with self.assertRaises(ValueError):
            sf.delete_inference("FAKE_ID")        

        # Remove existing inference
        s = sf.add_inference("TEST_inference_NAME")
        retrieved = sf.get_inference(s.get("id"))
        self.assertEqual(s, retrieved)

    def test_get_analyst(self):
        """
        TESTS: sadface.get_analyst()
        """
        sf.initialise()
        analyst = "A User"
        retrieved_analyst = sf.get_analyst()        
        self.assertEqual(retrieved_analyst, analyst)

    def test_get_atom(self):
        """
        TESTS: sadface.get_atom()
        """
        sf.initialise()
        self.assertEqual(sf.get_atom("unknown-id"), None)

        text = "DAKA DAKA"
        atom = sf.add_atom(text)
        atom_id = atom.get("id")
        result = sf.get_atom(atom_id)
        result_id = result.get("id")
        self.assertEqual(result_id, atom_id)

    def test_get_atom_id(self):
        """
        TESTS: sadface.get_atom_id()
        """
        sf.initialise()

        # Check behaviour when no atom to match against
        self.assertEqual(sf.get_atom_id("unknown-text"), None)

        # Add an atom, retrieve it by text content, and compare
        text = "DAKA DAKA"
        atom = sf.add_atom(text)
        atom_id = atom.get("id")
        retrieved_id = sf.get_atom_id(text)
        self.assertEqual(retrieved_id, atom_id)

    def test_get_atom_metadata(self):
        """
        TESTS: sadface.get_atom_metadata(atom_id, namespace=None, key=None)
        """
        sf.initialise()

        a = sf.add_atom("ATOM")
        sf.add_atom_metadata(a.get("id"), "TEST_NS", "TEST_KEY", "TEST_VAL")
        sf.add_atom_metadata(a.get("id"), "TEST_NS", "TEST_KEY2", "TEST_VA2")
        sf.add_atom_metadata(a.get("id"), "core", "TEST_KEY3", "TEST_VA3")

        expected = {'core': {'TEST_KEY3': 'TEST_VA3'}, 'TEST_NS': {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}}
        m = sf.get_atom_metadata(a.get("id"))
        self.assertEqual(expected, m)

        expected = {'TEST_KEY3': 'TEST_VA3'}
        m = sf.get_atom_metadata(a.get("id"), "core")
        self.assertEqual(expected, m)

        expected = {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}
        m = sf.get_atom_metadata(a.get("id"), "TEST_NS")
        self.assertEqual(expected, m)

        expected = "TEST_VA2"
        m = sf.get_atom_metadata(a.get("id"), "TEST_NS", "TEST_KEY2")
        self.assertEqual(expected, m)


    def test_get_atom_text(self):
        """
        TESTS: sadface.get_atom_text()
        """
        sf.initialise()

        # Check behaviour when no atom to match against
        self.assertEqual(sf.get_atom_text("unknown-id"), None)

        # Add an atom, retrieve it by text content, and compare
        text = "DAKA DAKA"
        atom = sf.add_atom(text)
        atom_id = atom.get("id")
        retrieved_text = sf.get_atom_text(atom_id)
        self.assertEqual(retrieved_text, text)


    def test_get_claim(self):
        """
        Tests: sadface.get_claim() with default values after init
        """
        sf.initialise()
        out = sf.get_claim()
        expected = None
        self.assertEqual(out, expected)

    def test_get_connections(self):
        """
        TESTS: sadface.get_connections()
        """
        sf.initialise()

        # Test with a non-existant ID
        results = sf.get_connections("TESTID")
        self.assertEqual([], results)

        # Test with existing ID that we create with no connections
        source_atom = sf.add_atom("SOURCE")
        results = sf.get_connections(source_atom.get("id"))
        self.assertEqual([], results)

        # Test with existing ID + known added connections
        target_atom = sf.add_atom("TARGET")
        edge = sf.add_edge(source_atom.get("id"), target_atom.get("id"))

        results = sf.get_connections(source_atom.get("id"))
        self.assertNotEqual([], results)
        self.assertEqual(len(results), 1)
        self.assertEqual(edge.get("id"), results[0].get("id"))
        self.assertEqual(source_atom.get("id"), results[0].get("source_id"))
        self.assertEqual(target_atom.get("id"), results[0].get("target_id"))


    def test_get_description(self):
        """
        Tests: sadface.get_description() with defaults values
        after init
        """
        sf.initialise()
        out = sf.get_description()
        expected = None
        self.assertEqual(out, expected)

    def test_get_document_id(self):
        """
        Tests: sadface.get_argument_id() with default values after init
        """
        sf.initialise()
        out = sf.get_document_id()
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

    def test_get_edge(self):
        """
        TESTS: sadface.get_edge()
        """
        sf.initialise()
        source = sf.add_atom("source")
        target = sf.add_atom("target")
        edge = sf.add_edge(source.get("id"), target.get("id"))

        retrieved_edge = sf.get_edge(edge.get("id"))
        self.assertEqual(retrieved_edge.get("source_id"), source.get("id"))
        self.assertEqual(retrieved_edge.get("target_id"), target.get("id"))
        self.assertEqual(retrieved_edge.get("id"), edge.get("id"))

    def test_get_global_metadata(self):
        """
        TESTS: sadface.get_global_metadata(namespace=None, key=None)
        """
        sf.initialise()

        m = sf.get_global_metadata()

        v = sf.get_version()
        docid = sf.get_document_id()

        self.assertEqual(v, m.get("core").get("version"))
        self.assertEqual(docid, m.get("core").get("id") )

        sf.add_global_metadata("TEST_NS", "TEST_KEY", "TEST_VAL")
        sf.add_global_metadata("TEST_NS", "TEST_KEY2", "TEST_VA2")

        expected = {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}
        m = sf.get_global_metadata("TEST_NS")
        self.assertEqual(expected, m)

        expected = "TEST_VA2"
        m = sf.get_global_metadata("TEST_NS", "TEST_KEY2")
        self.assertEqual(expected, m)

    def test_get_node(self):
        """
        TESTS: sadface.get_node()
        """
        sf.initialise()

        # Test with a non-existant ID
        result = sf.get_node("TEST")
        self.assertEqual(None, result)

        # Test with existin ID that we create
        atom = sf.add_atom("TESTATOM")
        result = sf.get_node(atom.get("id"))
        self.assertEqual(atom.get("id"), result.get("id"))

    def test_get_notes(self):
        """
        Tests: sadface.get_notes() with default values after init
        """
        sf.initialise()
        out = sf.get_notes()
        expected = None
        self.assertEqual(out, expected)

    def test_get_resource_metadata(self):
        """
        TESTS: sadface.get_atom_metadata(atom_id, namespace=None, key=None)
        """
        sf.initialise()

        r = sf.add_resource("RESOURCE")
        sf.add_resource_metadata(r.get("id"), "TEST_NS", "TEST_KEY", "TEST_VAL")
        sf.add_resource_metadata(r.get("id"), "TEST_NS", "TEST_KEY2", "TEST_VA2")
        sf.add_resource_metadata(r.get("id"), "core", "TEST_KEY3", "TEST_VA3")

        expected = {'core': {'TEST_KEY3': 'TEST_VA3'}, 'TEST_NS': {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}}
        m = sf.get_resource_metadata(r.get("id"))
        self.assertEqual(expected, m)

        expected = {'TEST_KEY3': 'TEST_VA3'}
        m = sf.get_resource_metadata(r.get("id"), "core")
        self.assertEqual(expected, m)

        expected = {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}
        m = sf.get_resource_metadata(r.get("id"), "TEST_NS")
        self.assertEqual(expected, m)

        expected = "TEST_VA2"
        m = sf.get_resource_metadata(r.get("id"), "TEST_NS", "TEST_KEY2")
        self.assertEqual(expected, m)

    def test_get_inference(self):
        """
        TESTS: sadface.get_inference() with default values
        """
        sf.initialise()
        result = sf.get_node("invalid-uuid")
        self.assertEqual(result, None)

    def test_get_inference_metadata(self):
        """
        TESTS: sadface.get_inference_metadata(inference_id, namespace=None, key=None)
        """
        sf.initialise()

        a = sf.add_inference("inference")
        sf.add_inference_metadata(a.get("id"), "TEST_NS", "TEST_KEY", "TEST_VAL")
        sf.add_inference_metadata(a.get("id"), "TEST_NS", "TEST_KEY2", "TEST_VA2")
        sf.add_inference_metadata(a.get("id"), "core", "TEST_KEY3", "TEST_VA3")

        expected = {'core': {'TEST_KEY3': 'TEST_VA3'}, 'TEST_NS': {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}}
        m = sf.get_inference_metadata(a.get("id"))
        self.assertEqual(expected, m)

        expected = {'TEST_KEY3': 'TEST_VA3'}
        m = sf.get_inference_metadata(a.get("id"), "core")
        self.assertEqual(expected, m)

        expected = {'TEST_KEY': 'TEST_VAL', 'TEST_KEY2': 'TEST_VA2'}
        m = sf.get_inference_metadata(a.get("id"), "TEST_NS")
        self.assertEqual(expected, m)

        expected = "TEST_VA2"
        m = sf.get_inference_metadata(a.get("id"), "TEST_NS", "TEST_KEY2")
        self.assertEqual(expected, m)

    def test_get_source(self):
        """
        TESTS: sadface.get_source()
        """
        sf.initialise()

        # Add an atom
        text = "DAKA DAKA"
        atom = sf.add_atom(text)
        atom_id =  sf.contains_atom(text)
        
         # Add a resource
        resource_text = "test resource"
        resource = sf.add_resource(resource_text)
        resource_id = resource.get("id")

        s = sf.get_source(atom_id, resource_id)
        self.assertEqual(s, None)


        # Add source to the atom, referencing the resource
        offset = 5
        length = len(resource_text)
        sf.add_source(atom_id, resource_id, resource_text, offset, length)

        # Now retrieve the source and test it
        s = sf.get_source(atom_id, resource_id)
        self.assertEqual(s.get("resource_id"), resource_id)
        self.assertEqual(s.get("text"), resource_text)
        self.assertEqual(s.get("offset"), offset)
        self.assertEqual(s.get("length"), length)

    def test_get_title(self):
        """
        Tests: sadface.get_title() with defaults values
        after init
        """
        sf.initialise()
        out = sf.get_title()
        expected = None
        self.assertEqual(out, expected)

    def test_get_version(self):
        """
        Tests: sadface.get_version()
        """
        sf.initialise()
        out = sf.get_version()
        expected = sf.version.__version__
        self.assertEqual(out, expected)

    def test_list_arguments(self):
        """
        Tests: sadface.get_arguments() with default values after init
        """
        sf.initialise()
        out = sf.list_arguments()
        expected = []
        self.assertEqual(out, expected)

    def test_list_atoms(self):
        """
        Tests: sadface.list_atoms() with default values after init
        """
        sf.initialise()
        out = sf.list_atoms()
        expected = []
        self.assertEqual(out, expected)

    def test_list_resources(self):
        """
        Tests: sadface.list_resources() 
        """
        sf.initialise()
        out = sf.list_resources()
        expected = []
        self.assertEqual(out, expected)

    def test_list_inferences(self):
        """
        Tests: sadface.get_arguments() with default values after init
        """
        sf.initialise()
        out = sf.list_inferences()
        expected = []
        self.assertEqual(out, expected)

    def test_get_created(self):
        """
        Tests: sadface.get_created()
        """
        sf.initialise()

        # Check that timestamp for creation time is a string
        timestamp = sf.get_created()
        self.assertEqual(True, isinstance(timestamp,str))

    def test_get_edited(self):
        """
        Tests: sadface.get_edited()
        """
        sf.initialise()

        # Check that timestamp for edited time is a string
        timestamp = sf.get_edited()
        self.assertEqual(True, isinstance(timestamp,str))

    def test_get_resource(self):
        """
        TESTS: sadface.get_resource(resource_id)
        """
        sf.initialise()
        new_resource = sf.add_resource("DAKA DAKA")
        new_resource_id = new_resource.get("id")
        
        retrieved_resource = sf.get_resource(new_resource_id)
        self.assertEqual(retrieved_resource, new_resource)
        

    def test_get_inference(self):
        """
        TESTS: sadface.get_inference() after an inference has been added
        """
        sf.initialise()
        inference_node = sf.add_inference("test-inference")
        inference_node_id = inference_node.get("id")

        result = sf.get_inference(inference_node_id)
        self.assertEqual(result.get("id"), inference_node_id)
        self.assertEqual(result.get("text"), inference_node.get("text"))

    def test_new_atom(self):
        """
        TESTS: sadface.new_atom()

        An atom dict should look like this:

            {"id":new_uuid(), "type":"atom", "text":text, "sources":[], "metadata":{}}

        So we check that it has the right keys and default values
        """
        text = "DAKA DAKA"
        atom = sf.new_atom(text)
        self.assertTrue(atom.get("id"))
        self.assertTrue(atom.get("type"))
        self.assertTrue(atom.get("type"), "atom")
        self.assertTrue(atom.get("text"))
        self.assertTrue(atom.get("text"), text)
        self.assertTrue(type(atom.get("sources")) is list)
        self.assertEqual(len(atom.get("sources")), 0)
        self.assertTrue(type(atom.get("metadata")) is dict)
        self.assertTrue(atom.get("metadata"))

    def test_new_edge(self):
        """
        TESTS: sadface.new_edge()

        An edge dict should look like this:

            {"id":new_uuid(), "source_id":source_id, "target_id":target_id}

        So we check that it has the right keys and default values
        """
        text = "DAKA DAKA"
        src_atom = sf.new_atom(text)
        src_id = src_atom.get("id")
        dest_atom = sf.new_atom(text)
        dest_id = dest_atom.get("id")

        edge = sf.new_edge(src_id, dest_id)

        self.assertTrue(edge.get("id"))
        out = edge.get("id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

        self.assertTrue(edge.get("source_id"))
        out = edge.get("source_id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

        self.assertTrue(edge.get("target_id"))
        out = edge.get("target_id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

    def test_new_resource(self):
        """
        TESTS: sadface.new_resource()
        """
        content = "DAKA DAKA"
        res = sf.new_resource(content)
        
        self.assertTrue(res.get("id"))
        out = res.get("id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

        self.assertTrue(res.get("content"))
        self.assertEqual(res.get("content"), content)

        self.assertTrue(res.get("type"))
        self.assertEqual(res.get("type"), "text")

        self.assertTrue(res.get("metadata"))
        self.assertTrue(type(res.get("metadata")) is dict)


    def test_new_sadface(self):
        """
        TESTS: sadface.new_sadface()
        """
        sd = sf.new_sadface()

        self.assertTrue(sd.get("metadata"))
        self.assertTrue(type(sd.get("metadata")) is dict)

        self.assertTrue(sd.get("metadata").get("core"))
        self.assertTrue(type(sd.get("metadata").get("core")) is dict)

        self.assertTrue(sd.get("metadata").get("core").get("version"))
        self.assertTrue(type(sd.get("metadata").get("core").get("version")) is str)

        self.assertTrue(sd.get("metadata").get("core").get("id"))
        self.assertTrue(type(sd.get("metadata").get("core").get("id")) is str)
        out = sd.get("metadata").get("core").get("id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)
        
        self.assertTrue(sd.get("metadata").get("core").get("analyst_name"))
        self.assertTrue(type(sd.get("metadata").get("core").get("analyst_name")) is str)
        
        self.assertTrue(sd.get("metadata").get("core").get("analyst_email"))
        self.assertTrue(type(sd.get("metadata").get("core").get("analyst_email")) is str)
        
        self.assertTrue(sd.get("metadata").get("core").get("created"))
        self.assertTrue(type(sd.get("metadata").get("core").get("created")) is str)

        self.assertTrue(sd.get("metadata").get("core").get("edited"))
        self.assertTrue(type(sd.get("metadata").get("core").get("edited")) is str)

        self.assertTrue(type(sd.get("resources")) is list)
        self.assertEqual(len(sd.get("resources")), 0)
        
        self.assertTrue(type(sd.get("nodes")) is list)
        self.assertEqual(len(sd.get("nodes")), 0)

        self.assertTrue(type(sd.get("edges")) is list)
        self.assertEqual(len(sd.get("edges")), 0)

    def test_new_inference(self):
        """
        TESTS: sadface.new_inference()
        """
        name = "DAKA"
        inf = sf.new_inference(name)

        self.assertTrue(inf.get("id"))
        out = inf.get("id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

        self.assertTrue(inf.get("type"))
        self.assertEqual(inf.get("type"), "inference")

        self.assertTrue(inf.get("name"))
        self.assertEqual(inf.get("name"), name)

        self.assertTrue(inf.get("metadata"))
        self.assertTrue(type(inf.get("metadata")) is dict)
    
    def test_new_source(self):
        """
        TESTS: sadface.new_source()
        {"resource_id":resource_id, "text":text, "offset":offset, "length":length}
        """
        test_id = sf.new_uuid()
        test_txt = "DAKA DAKA MORE DAKA"
        test_offset = 100
        src = sf.new_source(test_id, test_txt, test_offset)

        self.assertTrue(src.get("resource_id"))
        out = src.get("resource_id")
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

        self.assertTrue(src.get("text"))
        self.assertEqual(src.get("text"), test_txt)
    
        self.assertTrue(src.get("offset"))
        self.assertEqual(src.get("offset"), test_offset)

        self.assertTrue(src.get("length"))
        self.assertEqual(src.get("length"), len(test_txt))

    def test_new_uuid(self):
        """
        TESTS: sadface.new_uuid()
        """
        out = sf.new_uuid()
        result = False
        try:
            if UUID(out, version=4):
                result = True
        except:
            pass
        self.assertTrue(result)

    def test_now(self):
        """
        TESTS: sadface.now()
        """
        current = datetime.datetime.fromisoformat(sf.now())
        
        self.assertTrue(type(current) is datetime.datetime)

    def test_reset(self):
        """
        Tests: sadface.reset()

        A sadface document is created and manipulated then reset is used
        to return the document to it's initial state
        """

        # Iniitialise a SADFace document
        sf.initialise()
        expected = None
        out = sf.get_title()
        self.assertEqual(out, expected)

        # Explicitly alter it
        expected = "DAKA DAKA"
        sf.set_title(expected)
        out = sf.get_title()
        self.assertEqual(out, expected)

        # Reset the document - this should now be in the pre-init, empty-dict state
        sf.reset()
        expected = {}
        out = sf.sd
        self.assertEqual(out, expected)

    def test_save(self):
        """
        """
        pass

    def test_set_analyst(self):
        """
        TESTS: sadface.set_analyst(analyst_name)
        """
        sf.initialise()

        # TEST 1: Default analyst
        self.assertEqual(sf.get_analyst(), "A User")

        # TEST 2: Set user then retrieve & compare
        new_user = "Known User"
        sf.set_analyst(new_user)
        retrieved_user = sf.get_analyst()
        self.assertEqual(new_user, retrieved_user)

    def test_set_atom_text(self):
        """
        TESTS: sadface.set_atom_text()
        """
        sf.initialise()

        atom = sf.add_atom("DAKA DAKA")
        atom_id = atom.get("id")
        sf.set_atom_text(atom_id, "MORE DAKA")
        self.assertEqual(sf.get_atom_text(atom_id),"MORE DAKA")

    def test_set_claim(self):
        """
        TESTS: sadface.set_claim()
        """
        sf.initialise()
        self.assertEqual(sf.get_claim(), None)

        atom = sf.add_atom("DAKA DAKA")
        atom_id = atom.get("id")
        sf.set_claim(atom_id)
        self.assertEqual(sf.get_claim(), atom_id)

    def test_set_created(self):
        """
        TESTS: sadface.set_analyst(analyst_name)
        """
        sf.initialise()

        # TEST 1: Set created timestamp then retrieve & compare
        timestamp = sf.now()
        sf.set_created(timestamp)
        retrieved_timestamp = sf.get_created()
        self.assertEqual(timestamp, retrieved_timestamp)

    def test_set_description(self):
        """
        Tests: sadface.get_description() & set_description

        1. Set description of doc to known value
        2. Retrieve description & compare
        """
        sf.initialise()
        d = "test description"
        sf.set_description(d)
        out = sf.get_description()
        expected = d
        self.assertEqual(out, expected)

    def test_set_edited(self):
        """
        TESTS: sadface.set_edited(timestamp)
        """
        sf.initialise()

        # TEST 1: Set edited timestamp then retrieve & compare
        timestamp = sf.now()
        sf.set_edited(timestamp)
        retrieved_timestamp = sf.get_edited()
        self.assertEqual(timestamp, retrieved_timestamp)

    def test_set_title(self):
        """
        Tests: sadface.get_title() & set_title
        """
        sf.initialise()
        t = "test title"
        sf.set_title(t)
        out = sf.get_title()
        expected = t
        self.assertEqual(out, expected)

    def test_set_document_id(self):
        """
        TESTS: sadface.set_document_id(id)
        """
        sf.initialise()
        current_id = sf.get_document_id()
        self.assertNotEqual(None, current_id)

        test_id = "1234567890abcdefgh"
        sf.set_document_id(test_id)
        current_id = sf.get_document_id()
        self.assertEqual(test_id, current_id)

    def test_set_inference_name(self):
        """
        """
        sf.initialise()

        # Test setting non-existent inference
        with self.assertRaises(Exception) as context:
            sf.set_inference_name("TEST_ID", "TEST_NAME")
        out = str(context.exception)
        expected = 'Could not set the name of inference: TEST_ID'
        self.assertEqual(out, expected)

        test_name = "TEST_INFERENCE"
        s = sf.add_inference(test_name)
        retrieved = sf.get_inference(s.get("id"))
        self.assertEqual(s.get("name"), retrieved.get("name"))

        new_inference_name = "NEW_INFERENCE_NAME"
        sf.set_inference_name(s.get("id"), new_inference_name)
        retrieved = sf.get_inference(s.get("id"))
        self.assertEqual(new_inference_name, retrieved.get("name"))
        

    def test_verify(self):
        """
        """
        pass



if __name__ == "__main__":
    
    unittest.main()



#!/usr/bin/python

from . import aml
from . import sadface
from . import config
from . import validation
from . import version

import argparse
import cmd
import sys
import textwrap
import uuid

class REPL(cmd.Cmd):
    """
    The SADFace REPL. Type 'help' or 'help <command>' for assistance
    """

    def __init__(self):
        cmd.Cmd.__init__(self)
        self.prompt = '> '
        self.intro = "The SADFace REPL. Type 'help' or 'help <command>' for assistance"

    def do_arg(self, line):
        """
        Arguments are depicted in the following fashion e.g. premise1,premise2}conclusion
        premises are a comma separated list of strings where each string depicts a single
        premise. The conclusion is written at the end of the premise list using `}` to
        indicate a defeasible Modus Ponens rule. This uses the Simple Argument Description
        Notation (SADN).

        The line is split initially on the '}" to yield the premises (in the head)
        and the conclusion in the tail. The head is further split on the comma delimiters
        to retrieve each individual premise.
        """
        conid = None
        contxt = None
        if "}" in line:
            head,tail = line.split("}")
            if tail.startswith("id="):
                conid = tail.replace("id=", "")
            else:
                if (len(tail)>0):
                    contxt = tail.strip()
            premtext = []
            premid = []
            for element in head.split(","):
                t = element.strip()
                if t.startswith("id="):
                    premid.append(t.replace("id=", ""))
                else:
                    if (len(t) >0 ):
                        premtext.append(t)
            if ((conid is not None or contxt is not None) and (premtext or premid)):
                arg = sadface.add_argument(con_text=contxt, prem_text=premtext, con_id=conid, prem_id=premid)
                print(arg)
            else:
                print("USAGE: arg premise,premise,...}conclusion")
        else:
            print("USAGE: arg premise,premise,...}conclusion")

    def default(self, line):
        print("I do not understand that command. Type 'help' for a list of commands.")

    def do_add_resource(self, line):
        sadface.add_resource("hello world")
        print(sadface.prettyprint())

    def do_clear(self, line):
        sadface.sd = {}
        sadface.sd = sadface.initialise()

    def do_list_atoms(self, line):
        atoms = sadface.list_atoms()
        for a in atoms:
            print(a["id"]+" | "+a["text"])
    
    def do_print(self, line):
        print(sadface.sd)

    def do_prettyprint(self, line):
        print(sadface.prettyprint())

    def do_save(self, line):
        if('' != line):
            sadface.save(line)
        else:
            sadface.save()

    def do_save_to_dot(self, line):
        if('' != line):
            sadface.save(line, filetype="dot")
        else:
            sadface.save(filetype="dot")

    def do_quit(self, line):
        """
        Quit the SADRace REPL.
        """
        return True

    def do_version(self, line):
        print(version.__version__)

    def emptyline(self):
        pass

    def help_clear(self):
        print("Creates a new, empty SADFace document")


    do_p = do_print
    do_q = do_quit
    do_s = do_save
    do_sd = do_save_to_dot
    do_pp = do_prettyprint
 


def main():
    parser = argparse.ArgumentParser(description="This is the SADFace Python tool")
    parser.add_argument("-c", "--config", help="Supply a config file for SADFace to use.")
    parser.add_argument("-i", "--interactive", help="Use the SADFace REPL", action="store_true")

    fileinput = parser.add_mutually_exclusive_group()
    fileinput.add_argument("-l", "--load", help="Load a Sadface document from a file")
    fileinput.add_argument("-r", "--raw", help="Load a raw JSON document string into SADFace")
    fileinput.add_argument("--aml", help="Load a legacy AML document from a file")

    parser.add_argument("-s", "--save", help="Save the loaded document to a SADFace formatted JSON file")
    parser.add_argument("-v", "--verify", help="Verify that the supplied SADFace document is well formed")
    
    export = parser.add_mutually_exclusive_group()
    export.add_argument("--printdoc", 
        help="Print the SADFace document to the screen", action="store_true")
    export.add_argument("--pretty", 
        help="Pretty print the SADFace document", action="store_true")
    export.add_argument("--exportdot", 
        help="Export the SADFace document to dot format", action="store_true")

    parser.add_argument("--version", help="Display the current version of the SADFace library", action="store_true")

    args = parser.parse_args()

    if args.config:
        config.set_config_location(args.config)

    if args.raw:
        sadface.sd = sadface.import_json(args.raw)
    elif args.load:
        sadface.sd = sadface.load_from_file(args.load)
    elif args.aml:
        aml.init(args.aml)
    else:
        sadface.sd = sadface.initialise()

    if args.save:
        sadface.save(args.save)

    if args.verify:
        sadface.sd = sadface.load_from_file(args.verify)
        if sadface.sd:
            result = validation.verify(sadface.sd, False)
            print(result)

    if args.printdoc:
        print(sadface.print_doc())
    elif args.pretty:
        print(sadface.prettyprint())
    elif args.exportdot:
        print(sadface.export_dot())
    
    if args.version:
        print(version.__version__)
    
    if args.interactive:
        REPL().cmdloop()
        
            



#!/usr/bin/python

import uuid

from . import sadface


def check_edges_block(doc):
    """
    The edges block is a list of edges within the SADFace document and is a required
    part of a well formed SADFace document
    """
    problems = []

    if doc.get("edges") == None:
        problems.append("No 'edges' block")
    else:
        if type(doc.get("edges")) is not list:
            problems.append("Edges block is not a list")
        for edge in doc.get("edges"):
            if type(edge) is not dict:
                problems.append("Edges block has a member that is not a dict/object")

            if "id" not in edge:
                problems.append("Edge has no 'id' key")
            else:
                edge_id = edge.get("id")
                try:
                    uuid.UUID(edge_id, version=4)
                except:
                    problems.append("'edge_id' is not an instance of UUID4") 
                    
            if "source_id" not in edge:
                problems.append("Edge has no 'source_id' key")
            else:
                source_id = edge.get("source_id")
                try:
                    uuid.UUID(source_id, version=4)
                except:
                    problems.append("'source_id' is not an instance of UUID4")

            if "target_id" not in edge:
                problems.append("Edge has no 'target_id' key")
            else:
                target_id = edge.get("target_id")
                try:
                    uuid.UUID(target_id, version=4)
                except:
                    problems.append("'target_id' is not an instance of UUID4")

            for key in edge.keys():
                if key not in ["id", "source_id", "target_id"]:
                    problems.append("Edge contains the invalid key: "+key)
                    
    return problems

def check_global_metadata_block(doc):
    """
    The global metadata block is a collection of objects, each of which contains information
    about the current document. A core collection is a required part of the global metadata
    block and is reserved for SADFace metadata only. Additional metadata can be stored in
    user defined metadata blocks that are named uniquely and store key:value pairs.

    Global metadata is a required part of a well formed SADFace document

    """
    problems = []

    if doc.get("metadata") == None:
        problems.append("No 'metadata' block")
    else:
        if type(doc.get("metadata")) is not dict:
            problems.append("Metadata block is not a dict/object")
        else:
            if doc.get("metadata").get("core") is None:
                problems.append("No 'core' block in metadata")
            else:
                core = doc.get("metadata").get("core")
                if type(core) is not dict:
                    problems.append("Metadata core block is not a dict/object")
                
                if "analyst_email" not in core:
                    problems.append("'analyst_email' key is not in the core metadata block")
                else:
                    analyst_email = core.get("analyst_email")
                    if type(analyst_email) is not str:
                        problems.append("'analyst_email' in core metadata is not a string")

                if "analyst_name" not in core:
                    problems.append("'analyst_name' key is not in the core metadata block")
                else:
                    analyst_name = core.get("analyst_name")
                    if type(analyst_name) is not str:
                        problems.append("'analyst_name' in core metadata is not a string")

                if "created" not in core:
                    problems.append("'created' key is not in the core metadata block")
                else:
                    created = core.get("created")
                    if type(created) is not str:
                        problems.append("'created' in core metadata is not a string")
                
                if "edited" not in core:
                    problems.append("'edited' key is not in the core metadata block")
                else:
                    edited = core.get("edited")
                    if type(edited) is not str:
                        problems.append("'edited' in core metadata is not a string")

                if "id" not in core:
                    problems.append("'id' key is not in the core metadata block")
                else:
                    core_id = core.get("id")
                    if type(core_id) is not str:
                        problems.append("'id' in core metadata is not a string")

                if "version" not in core:
                    problems.append("'version' key is not in the core metadata block")
                else:
                    version = core.get("version")
                    if type(version) is not str:
                        problems.append("'version' in core metadata is not a string")

                for key in core.keys():
                    if key not in ["analyst_email", "analyst_name", "created", "edited", "id", "version"]:
                        problems.append("Metadata contains the invalid key: "+key)
            
            for block in doc.get("metadata"):
                obj = doc.get("metadata").get(block)
                if type(obj) is not dict:
                    problems.append("Metadata contains a block that is not a dict/object:"+str(block))

    return problems

def check_node_type_atom(atom):
    """
    Atoms are a specific type of node in a SADFace graph. They have an associated set of data that
    is required for an Atom to be considered well formed.
    """
    problems = []

    if "text" not in atom:
        problems.append("No 'text' key in Atom Node")
    else:
        text = atom.get("text")
        if type(text) is not str:
            problems.append("'text' in Atom Node is not a string")

    if "sources" not in atom:
        problems.append("No 'text' key in Atom Node")
    else:
        text = atom.get("text")
        if type(text) is not str:
            problems.append("'text' in Atom Node is not a string")
    
    return problems

def check_node_type_scheme(scheme):
    """
    Schemes are a specific type of node in a SADFace graph. They have an associated set of data that
    is required for a Scheme to be considered well formed.

    """
    problems = []

    if "name" not in atom:
        problems.append("No 'name' key in Scheme Node")
    else:
        name = atom.get("name")
        if type(name) is not str:
            problems.append("'name' in Atom Node is not a string")
    
    return problems

def check_nodes_block(doc):
    """
    Nodes hold data within a SADFace graph and are connected by Edges. Nodes have a common associated 
    set of data that is required for them to be considered well formed. Several additional sub-types 
    of Node exist and may require additional data associated with them.

    """
    problems = []
    
    if doc.get("nodes") == None:
        problems.append("No 'nodes' block")
    else:
        if type(doc.get("nodes")) is not list:
            problems.append("Nodes block is not a list")
        else:
            for node in doc.get("nodes"):
                if "id" not in node:
                    problems.append("Node doesn't contain an 'id' key")
                else:
                    node_id = node.get("id")
                    if type(node_id) is not str:
                        problems.append("Node 'id' key is not a string")
                    try:
                        uuid.UUID(node_id, version=4)
                    except:
                        problems.append("'node_id' is not an instance of UUID4"+str(node_id))

                if "type" not in node:
                    problems.append("Node doesn't contain a 'type' key")
                else:
                    type_key = node.get("type")
                    if type(type_key) is not str:
                        problems.append("Node 'type' key is not a string")
                    if type_key not in ["atom", "scheme"]:
                        problems.append("Node 'type' not of supported SADFace node type")

                    if type(type_key) == "atom":
                        problems += check_node_type_atom(node)
                    elif type(type_key) == "scheme":
                        problems += check_node_type_scheme(node)
                        
                metadata = node.get("metadata")
                if metadata:
                    for block in metadata:
                        obj = doc.get("metadata").get(block)
                        if type(obj) is not dict:
                            problems.append("Metadata contains a block that is not a dict/object:"+str(obj))
                        else:
                            if doc.get("metadata").get("core") is None:
                                problems.append("No 'core' block in metadata")

        for block in doc.get("nodes"):
            if type(block) is not dict:
                problems.append("Nodes contains a block that is not a dict/object:"+str(block))
            

    return problems
 

def check_resources_block(doc):
    """
    The resources block is a list of, possibly external, resources that the SADFace graph refers to.
    The block is required but may be an empty list.
    """
    problems = []
    
    if doc.get("resources") == None:
        problems.append("No 'resources' block")
    else:
        if type(doc.get("resources")) is not list:
            problems.append("Resources block is not a list")

        for block in doc.get("resources"):
            if type(block) is not dict:
                problems.append("Resources contains a block that is not a dict/object:"+str(block))

    return problems



def verify(incoming=None, as_string=False):
    """
    Verifies that the supplied SADFace document is well formed.

    If no input is supplied then this function will verify the current internal sadface document.
    Otherwise a document can be supplied in two forms, as a JSON encoded string or as a Python dict.
    In either case the function checks the input type and acts accordingly to normalise the input to
    a Python dict. This dict is then checked to ensure that:

        1. All necessary keys are present
        2. All supplied keys are permitted, i.e. within the boundaries permitted
            by the namespacing/metadata infrastructure
        3. No additional keys are present
        4. All values, where appropriate, are within defined parameters

    Returns: a pair (status, problems) such that:
        status is True if their are no problems and False otherwise
        problems is either a list of strings where each string describes a problem with the supplied input or a single string in which the members of the list have been joined (useful for convenient output to the user)
    """
    doc = {}
    problems = []

    if incoming is not None:
        if type(incoming) is str:
            doc = sadface.import_json(incoming)
        elif type(incoming) is dict:
            doc = incoming
        else:
            problems.append("Supplied document is neither a JSON encoded string nor a Python dict")
            return problems
    else:
        doc = sadface.sd


    problems += check_global_metadata_block(doc)
    problems += check_edges_block(doc)
    problems += check_nodes_block(doc)
    problems += check_resources_block(doc)
    

    if len(problems) > 0 and as_string == False:
        return False, problems
    elif len(problems) == 0 and as_string == False:
        return True, problems
    elif len(problems) > 0 and as_string == True:
        return False, ", ".join(problems)
    else:
        return True, ", ".join(problems)
    




#!/usr/bin/python

import argparse
import cmd
import codecs
import datetime
import json
import sys
import textwrap
import uuid

from . import config
from . import validation
from . import version

sd = {}

def add_support(con_text=None, prem_text=None, con_id=None, prem_id=None):
    """
    Syntactic sugar to create an argument structure from a set of texts.
    Given a conclusion text & a list of premise texts. Creates an intermediate,
    default "support" scheme.

    This makes it easier to build a SADFace document without manually creating
    and organising individual nodes.

    Returns an argument dict, e.g.

    {
        "conclusion": atom,
        "scheme": atom,
        "premises": [atom(s)]
    }

    Returns: a dict
    """
    if((con_text is not None or con_id is not None) and (prem_text is not None or prem_id is not None)):

        if con_text is not None:
            c = add_atom(con_text)
        else:
            c = get_atom(con_id)

        s = add_inference("support")
        try:
            add_edge(s["id"], c["id"])
        except Exception as ex:
            print(ex)
            raise Exception("Could not create new argument")

        p_list = []
        if(prem_text is not None):
            for text in prem_text:
                atom = add_atom(text)
                p_list.append(atom["id"])
                try:
                    add_edge(atom["id"], s["id"])
                except Exception as ex:
                    print(ex)
                    raise Exception("Could not create new argument")
        if(prem_id is not None):
            for atom_id in prem_id:
                atom = get_atom(atom_id)
                p_list.append(atom["id"])
                try:
                    add_edge(atom["id"], s["id"])
                except Exception as ex:
                    print(ex)
                    raise Exception("Could not create new argument")

        arg = {"conclusion":c, "scheme":s, "premises":p_list}
        return arg
    return None


def add_edge(source_id, target_id):
    """
    Given a source atom ID & a target atom ID, create an 
    edge linking the two and add it to the sadface doc,
    "sd" & return the dict representing the edge. If
    either of source or target IDs is invalid then an
    exception is raised.

    Returns: a dict 
    """
    if ((get_node(source_id) is not None) and (get_node(target_id) is not None)):
        edge = new_edge(source_id, target_id)
        sd["edges"].append(edge)
        return edge
    raise Exception("Could not create new edge between: "+source_id+" & "+target_id)

def add_atom(text):
    """
    Create a new argument atom using the supplied text

    Returns: the new atom dict
    """
    atomid = contains_atom(text)
    atom = None

    if atomid is not None:
        atom = get_atom(atomid)

    else:
        atom = new_atom(text)
        sd["nodes"].append(atom)
    
    return atom

def add_atom_metadata(atom_id, namespace, key=None, value=None):
    """
    Add metadata, a key:value pair, to the atom dict identified
    by the supplied atom ID and metadata namespace.

    Note that "core" should be reserved for sadface default metadata only

    Additional namespaces can be any string but should normally follow
    the reversed domain name approach to provide some level of uniqueness
    and avoid unnecessary clashes
    """
    for node in sd["nodes"]:
        if "atom" == node["type"]:
            if atom_id == node["id"]:
                if node["metadata"].get(namespace) is None:
                    node["metadata"][namespace] = {}
                    node["metadata"][namespace][key] = value
                else:
                    node["metadata"][namespace][key] = value

def add_conflict(name):
    """
    Add a new conflict node dict to the sadface document. The conflict type
    is identified by the supplied name

    Returns: The new conflict dict
    """
    conflict = new_conflict(name)
    sd["nodes"].append(conflict)
    return conflict

def add_conflict_metadata(conflict_id, namespace, key=None, value=None):
    """
    Add metadata, a key:value pair, to the conflict dict identified
    by the supplied conflict ID and metadata namespace.

    Note that "core" is reserved for sadface default metadata only

    Additional namespaces can be any string but should normally follow
    the reversed domain name approach to provide some level of uniqueness
    and avoid unnecessary clashes
    """
    for node in sd["nodes"]:
        if "conflict" == node["type"]:
            if conflict_id == node["id"]:
                if node["metadata"].get(namespace) is None:
                    node["metadata"][namespace] = {}
                    node["metadata"][namespace][key] = value
                else:
                    node["metadata"][namespace][key] = value

def add_global_metadata(namespace, key, value):
    """
    Add metadata, a key:value pair to the base sadface doc
    """
    if sd["metadata"].get(namespace) is None:
        sd["metadata"][namespace] = {}
        sd["metadata"][namespace][key] = value
    else:
        sd["metadata"][namespace][key] = value

def add_notes(text):
    """
    Add a metadata entry for the document that contains notes. Notes
    are miscellaneous, unstructured free text.
    """
    sd["metadata"]["core"]["notes"] = text


def add_resource(content):
    """
    Create a new resource dict using the supplied content string
    then add to the resourses list of the sadface doc

    Returns: the new resource dict
    """
    res = new_resource(content)
    sd["resources"].append(res)
    return res

def add_resource_metadata(resource_id, namespace, key=None, value=None):
    """
    Add metadata, a key:value pair to the resource dict identified
    by the supplied atom ID.    
    """
    for res in sd["resources"]:
        if res["id"] == resource_id:
            if res["metadata"].get(namespace) is None:
                res["metadata"][namespace] = {}
                res["metadata"][namespace][key] = value
            else:
                res["metadata"][namespace][key] = value

def add_inference(name):
    """
    Add a new inference node dict to the sadface document. The scheme type
    associated with the inference is identified by the supplied name

    Returns: The new inference dict
    """
    inference = new_inference(name)
    sd["nodes"].append(inference)
    return inference

def add_inference_metadata(inference_id, namespace, key=None, value=None):
    """
    Add metadata, a key:value pair, to the inference dict identified
    by the supplied inference ID and metadata namespace.

    Note that "core" is reserved for sadface default metadata only.

    Additional namespaces can be any string but should normally follow
    the reversed domain name approach to provide some level of uniqueness
    and avoid unnecessary clashes.
    """
    for node in sd["nodes"]:
        if "inference" == node["type"]:
            if inference_id == node["id"]:
                if node["metadata"].get(namespace) is None:
                    node["metadata"][namespace] = {}
                    node["metadata"][namespace][key] = value
                else:
                    node["metadata"][namespace][key] = value


def add_source(atom_id, resource_id, text, offset, length):
    """
    Add a new source dict to the atom identified by the supplied
    atom ID. The new source refers to the an existing resource that
    is identified by the supplied resource ID. The source identifies
    text string in the resource dict that it references as well as
    the offset & length of the text from the beginning of the resource

    Returns: The new source dict
    """
    source = new_source(resource_id, text, offset)
    for node in sd["nodes"]:
        if "atom" == node["type"]:
            if atom_id == node["id"]:
                node["sources"].append(source)
                return source

def append_notes(text):
    """
    Append new text to an existing notes entry
    """
    if sd.get("metadata").get("core").get("notes") != None:
        sd["metadata"]["core"]["notes"] += text
    else:
        add_notes(text)

def build_argument(con_text=None, prem_text=None, con_id=None, prem_id=None):
    """
    Syntactic sugar to create an argument structure from a set of texts.
    Given a conclusion text & a list of premise texts. Creates an intermediate,
    default, "inference" node named "support".

    This makes it easier to build a SADFace document without manually creating
    and organising individual nodes.

    Returns an argument dict, e.g.

    {
        "conclusion": atom,
        "inference": atom,
        "premises": [atom ID(s)]
    }

    e.g.

    {
        "conclusion": {
            "id": "329aee40-b34d-4eaa-9e25-9ef3098b4cfa",
            "type": "atom",
            "text": "You should treasure every moment",
            "sources": [],
            "metadata": {
                "core": {}
            }
        },
        "inference": {
            "id": "f4ae4161-f18b-49f7-9213-d81c2084cedd",
            "type": "inference",
            "name": "support",
            "metadata": {
                "core": {}
            }
        },
        "premises": ["6b4cbe35-4e11-4126-ab25-97c9ac651647", "4cebed0d-3869-4a20-958c-9e164a019573"]
    }

    Returns: a dict
    """
    if((con_text is not None or con_id is not None) and (prem_text is not None or prem_id is not None)):

        if con_text is not None:
            c = add_atom(con_text)
        else:
            c = get_atom(con_id)

        i = add_inference("support")
        try:
            add_edge(i["id"], c["id"])
        except Exception as ex:
            print(ex)
            raise Exception("Could not create new argument")

        p_list = []
        if(prem_text is not None):
            for text in prem_text:
                atom = add_atom(text)
                p_list.append(atom["id"])
                try:
                    add_edge(atom["id"], i["id"])
                except Exception as ex:
                    print(ex)
                    raise Exception("Could not create new argument")
        if(prem_id is not None):
            for atom_id in prem_id:
                atom = get_atom(atom_id)
                p_list.append(atom["id"])
                try:
                    add_edge(atom["id"], i["id"])
                except Exception as ex:
                    print(ex)
                    raise Exception("Could not create new argument")

        arg = {"conclusion":c, "inference":i, "premises":p_list}
        return arg
    return None

def build_attack(attacker_text=None, target_text=None, attacker_id=None, target_id=None):
    """
    SADFace primarily support two forms of conflict, Disagreement and attack (We consider defeat
    to be a specific interpretation derived from the extistence of an attack relationship so do
    not explicity represent it at present. Defeat might be better represented as a status, perhaps 
    transient, that is either derived at runtime from the state of the graph or else is stored 
    more permanently within the metadata for a given attack). 

    A single attack node is used to point, via edge nodes from one atom node, the source of the 
    attack, to another atom node, the destination of the attack. Attack is a binary relationship 
    that has exactly two forms within SADFace:

    -- An argument may attack itself, i.e. the source and destination ends of the attack relate
    to the same atom.
    -- An argument may attack another argument, i.e. the source and destination are different
    atoms.

    We depict attack through the use of a node that represent the attack relationship. This
    function will instantiate a conflict node between two nodes (either pre-existing & identifed by 
    node IDs or created by this function from supplied texts, or a mixture of the two).

    Returns a conflict dict, e.g.

    {
        "attacker": atom,
        "conflict": atom,
        "target": atom
    }
    (where the scheme just happens to depict a conflict)

    
    For example:

    {
        "attacker": {
            "id": "6399c86d-3c07-4223-ae91-bb989b6dc21e",
            "type": "atom",
            "text": "You are going to die",
            "sources": [],
            "metadata": {
                "core": {}
            }
        },
        "conflict": {
            "id": "7b4b7d23-7386-470e-ad48-56a655812218",
            "type": "conflict",
            "name": "attack",
            "metadata": {
                "core": {}
            }
        },
        "target": {
            "id": "777a21cf-6c60-44cc-9b3f-bfe1f2e342d6",
            "type": "atom",
            "text": "I might live forever",
            "sources": [],
            "metadata": {
                "core": {}
            }
        }
    }

    Returns: a dict
    """
    if((attacker_text is not None or attacker_id is not None) and (target_text is not None or target_id is not None)):
        
        if attacker_text is not None:
            a = add_atom(attacker_text)
        else:
            a = get_atom(attacker_id)

        c = add_conflict("attack")

        try:
            add_edge(c["id"], a["id"])
        except Exception as ex:
            print(ex)
            raise Exception("Could not create new argument")

        if target_text is not None:
            d = add_atom(target_text)
        else:
            d = get_atom(target_id)

        try:
            add_edge(d["id"], c["id"])
        except Exception as ex:
            print(ex)
            raise Exception("Could not create new argument")

        arg = {"attacker":a, "conflict":c, "target":d}
        return arg
    return None


def build_disagreement(arg_text=None, disagreement_text=None, arg_id=None, disagreement_id=None):
    """
    SADFace primarily support two forms of conflict, Disagreement and attack (We consider defeat
    to be a specific interpretation derived from the extistence of an attack relationship so do
    not explicity represent it at present). Disagreements create a relationship between two or 
    more arguments to indicate that the arguments mutually disagree with each other.

    A single disagreement node is used to point, via edge nodes to the collection of arguments
    that disagree with each other. A disagreement must link at least two different argument nodes,
    identifed by different node IDs, in order to create the relationship between them.


    We depict conflict through the use of a node that represent the conflict relationship. This
    function will instantiate a conflict node between two nodes (either pre-existing & identifed by 
    node IDs or created by this function from supplied texts, or a mixture of the two).

    Returns a conflict dict, e.g.

    {
        "argument_1": atom,
        "conflict": atom,
        "argument_2": atom
    }
    (where the scheme just happens to depict a conflict)

    
    For example:

    {
        "argument_1": {
            "id": "6399c86d-3c07-4223-ae91-bb989b6dc21e",
            "type": "atom",
            "text": "You are going to die",
            "sources": [],
            "metadata": {
                "core": {}
            }
        },
        "conflict": {
            "id": "7b4b7d23-7386-470e-ad48-56a655812218",
            "type": "conflict",
            "name": "disagreement",
            "metadata": {
                "core": {}
            }
        },
        "argument_2": {
            "id": "777a21cf-6c60-44cc-9b3f-bfe1f2e342d6",
            "type": "atom",
            "text": "I might live forever",
            "sources": [],
            "metadata": {
                "core": {}
            }
        }
    }

    Returns: a dict
    """
    if((arg_text is not None or arg_id is not None) and (disagreement_text is not None or disagreement_id is not None)):
        
        if arg_text is not None:
            a = add_atom(arg_text)
        else:
            a = get_atom(arg_id)

        c = add_conflict("disagree")

        try:
            add_edge(c["id"], a["id"])
        except Exception as ex:
            print(ex)
            raise Exception("Could not create new argument")

        if disagreement_text is not None:
            d = add_atom(disagreement_text)
        else:
            d = get_atom(disagreement_id)

        try:
            add_edge(c["id"], d["id"])
        except Exception as ex:
            print(ex)
            raise Exception("Could not create new argument")

        arg = {"argument_1":a, "conflict":c, "argument_2":d}
        return arg
    return None


def clear_notes():
    """
    Remove any existing notes
    """
    sd.get("metadata").get("core").pop("notes")


def contains_atom(atom_text):
    """
    Searches the sadface document for an existing atom containing
    the supplied text. If found, returns the id of that atom,
    otherwise None
    """
    for node in sd["nodes"]:
        if "atom" == node["type"]:
            if atom_text == node["text"]:
                return node["id"]
    return None

def delete_atom(atom_id):
    """
    Remove the atom from the sadface document identified by the
    supplied atom ID
    """
    atom = get_atom(atom_id)
    sd["nodes"].remove(atom)

    conns = get_connections(atom_id)
    for c in conns:
        delete_edge(c["id"])

def delete_conflict(conflict_id):
    """
    Remove the conflict from the sadface document identified by the
    supplied conflict ID

    """
    conflict = get_conflict(conflict_id)
    sd["nodes"].remove(conflict)

    conns = get_connections(conflict_id)
    for c in conns:
        delete_edge(c["id"])

def delete_edge(edge_id):
    """
    Remove the edge from the sadface document identified by the
    supplied edge ID

    """
    edge = get_edge(edge_id)
    sd["edges"].remove(edge)

def delete_source(atom_id, resource_id):
    """
    Remove a source from the atom identified by the
    supplied atom ID & resource ID respectively

    """
    source = get_source(atom_id, resource_id)
    if source is not None:
        atom = get_atom(atom_id)
        atom["sources"].remove(source)

def delete_resource(resource_id):
    """
    Remove the resource from the sadface document identified by the
    supplied resource ID

    """
    resource = get_resource(resource_id)
    sd["resources"].remove(resource)

def delete_inference(inference_id):
    """
    Remove the inference from the sadface document identified by the
    supplied inference ID

    """
    inference = get_inference(inference_id)
    sd["nodes"].remove(inference)

    conns = get_connections(inference_id)
    for c in conns:
        delete_edge(c["id"])

def export_cytoscape():
    """
    Cytoscape.js is a useful graph visualisation library for Javascript. However
    it uses some slightly different keynames and includes description of visual
    elements, useful to Cytoscape's visualisation, but having no place in SADFace.

    Both nodes & edges in a Cytoscape graph are collated together into a single
    eleents object so we need to do that to the SADFace nodea & edges. Furthemore,
    each node and edge object must contain a data object. After that conversion is
    a relatively straightforward mapping:

    EDGES
        id -> id
        source_id -> source
        target_id -> target

        e.g. 
        {
            "data": {
                "source": "a1",
                "id": "a1s1",
                "target": "s1"
            }
        }

    NODES - ATOMS    
        id -> id
        type -> type
        text -> content
        + "classes":"atom-label"
        + "typeshape":"roundrectangle"

        e.g.
        {
            "classes": "atom-label",
            "data": {
                "content": "Every person is going to die",
                "type": "atom",
                "id": "a1",
                "typeshape": "roundrectangle"
            }
        }


    NODES - SCHEMES
        id -> id
        type -> type
        name -> content
        + "classes":"scheme-label"
        + "typeshape":"diamond"
        
        e.g.
        {
            "classes": "scheme-label",
            "data": {
                "content": "Default\nSupport",
                "type": "scheme",
                "id": "s1",
                "typeshape": "diamond"
            }
        }

    """
    cy = {}
    cy['elements'] = {}
    cy['elements']['nodes'] = []
    cy['elements']['edges'] = []

    for edge in sd['edges']:
        e = {}
        e['data'] = {}
        e['data']['id'] = edge['id']
        e['data']['source'] = edge['source_id']
        e['data']['target'] = edge['target_id']

        cy['elements']['edges'].append(e)

    for node in sd['nodes']:
        n = {}
        n['data'] = {}
        n['data']['id'] = node['id']
        n['data']['type'] = node['type']
        if n['data']['type'] == "atom":
            n['classes'] = "atom-label"
            n['data']['typeshape'] = "roundrectangle"
            n['data']['content'] = node['text']

        else:
            n['classes'] = "scheme-label"
            n['data']['typeshape'] = "diamond"
            n['data']['content'] = node['name']

        cy['elements']['nodes'].append(n)

    return  json.dumps(cy)

def export_dot(trad=True):
    """
    Exports a subset of SADFace to the DOT graph description language

    Returns: String-encoded DOT document
    """
    if trad:
        colour_scheme = "X11"
        support_colour = "darkolivegreen3"
        disagree_colour = "firebrick2"
        default_colour = "cornsilk4"
    else:
        colour_scheme = "ylgnbu3"
        support_colour = "1"
        disagree_colour = "3"
        default_colour = "2"

    max_length = 25
    edge_str = " -> "
    dot = "digraph SADFace {"
    dot += "node [style=\"filled\"]"
    for node in sd["nodes"]:
        if "text" in node:
            txt = node["text"]
            if len(txt) > max_length:
                txt = "\\n".join(textwrap.wrap(txt, max_length))
            line = '"{}"'.format(node['id']) + " [label=\"" + txt + "\"]" + " [shape=box, style=rounded];\n"
            dot += line
        elif "name" in node:
            if "support" == node.get("name"):
                line = '"{}"'.format(node['id']) + " [label=\"" + node["name"]\
                + "\"]"\
                + " ["\
                + "colorscheme="+colour_scheme+", fillcolor="+support_colour\
                + ", shape=diamond];\n"
            elif "disagree" == node.get("name"):
                line = '"{}"'.format(node['id']) + " [label=\"" + node["name"]\
                + "\"]"\
                + " ["\
                + "colorscheme="+colour_scheme+", fillcolor="+disagree_colour\
                + ", shape=diamond];\n"
            elif "attack" == node.get("name"):
                line = '"{}"'.format(node['id']) + " [label=\"" + node["name"]\
                + "\"]"\
                + " ["\
                + "colorscheme="+colour_scheme+", fillcolor="+disagree_colour\
                + ", shape=diamond];\n"
            else:
                line = '"{}"'.format(node['id']) + " [label=\"" + node["name"]\
                + "\"]"\
                + " ["\
                + "colorscheme="+colour_scheme+", fillcolor="+default_colour\
                + ", shape=diamond];\n"

            dot += line

    for edge in sd["edges"]:
        source = get_node(edge["source_id"])
        target = get_node(edge["target_id"])
        
        if("atom" == source["type"]):
            dot += '"{}"'.format(source["id"])
        elif "inference" == source["type"]:
            dot += '"{}"'.format(source["id"])
        elif "conflict" == source["type"]:
            dot += '"{}"'.format(source["id"])
        
        dot += edge_str

        if("atom" == target["type"]):
            dot += '"{}"'.format(target["id"])
        elif "inference" == target["type"]:
            dot += '"{}"'.format(target["id"])
        elif "conflict" == target["type"]:
            dot += '"{}"'.format(target["id"])

        
        dot += ";\n"

    dot += "}"
    
    return dot

def export_json():
    """
    Dump the current sadface document to a JSON string

    Returns: String-encoded JSON
    """
    return json.dumps(sd)

def get_analyst():
    """
    Retrieve the name of the argument analyst in the SADFace doc
    """
    return sd["metadata"]["core"]["analyst_name"]

def get_atom(atom_id):
    """
    Retrieve the atom dict identified by the supplied atom ID

    Returns: An atom dict
    """
    for node in sd["nodes"]:
        if atom_id == node["id"]:
            return node

def get_atom_id(text):
    """
    Retrieve the first atom whose text equals the supplied text

    Returns: The atom's ID or None 
    """
    for node in sd["nodes"]:
        if text == node.get("text"):
            return node["id"]

def get_atom_metadata(atom_id, namespace=None, key=None):
    """
    Retrieve an atom's metadata. Results depend upon the specificity of the request
    which must include atom_id.

    If atom_id only is provided then all metadata assocaited with the atom is returned
    If atom_id + namespace is provided then only the specified namespace is returned
    If atom_id + namespace + a key is provided then the value associated with that key is returned
    Else nothing is returned

    Returns: A metadata dict
    """
    atom = get_atom(atom_id)
    if atom is not None:
        if namespace is None:
            return atom.get("metadata")
        else:
            if key is None:
                return atom.get("metadata").get(namespace)
            else:
                return atom.get("metadata").get(namespace).get(key)


def get_atom_text(atom_id):
    """
    Retrieve the atom dict identified by the supplied atom ID

    Returns: An atom dict
    """
    for node in sd["nodes"]:
        if atom_id == node["id"]:
            return node["text"]

def get_claim():
    """
    Retrieve the claim metadata entry from the document
    """
    return sd["metadata"].get("core").get("claim")

def get_conflict(conflict_id):
    """
    Retrieve the conflit dict identified by the supplied conflict ID

    Returns: A conflict dict
    """
    for node in sd["nodes"]:
        if conflict_id == node["id"]:
            return node

def get_conflict_metadata(conflict_id, namespace=None, key=None):
    """
    Retrieve a conflict's metadata. Results depend upon the specificity of the request
    which must include a valid conflict_id.

    If conflict_id only is provided then all metadata associated with the conflict is returned
    If conflict_id + namespace is provided then only the specified namespace is returned
    If conflict_id + namespace + a key is provided then the value associated with that key is returned
    Else nothing is returned

    Returns: A metadata dict or None
    """
    conflict = get_conflict(conflict_id)
    if conflict is not None:
        if namespace is None:
            return conflict.get("metadata")
        else:
            if key is None:
                return conflict.get("metadata").get(namespace)
            else:
                return conflict.get("metadata").get(namespace).get(key)


def get_connections(node_id):
    """
    Given a node id, retrieve a list containing all edges that connnect it
    to other nodes
    """
    conn =  []
    for edge in sd["edges"]:
        if node_id == edge["source_id"] or node_id == edge["target_id"]:
            conn.append(edge)
    return conn

def get_created():
    """
    Retrieve the claim metadata entry from the document
    """
    return sd["metadata"].get("core").get("created")

def get_document():
    """
    Retrieve the current SADFace document
    """
    return sd

def get_document_id(doc = None):
    """
    Retrieve the SADFace document's ID
    """
    if doc is None:
        return sd["metadata"].get("core").get("id")
    else:
        return doc["metadata"].get("core").get("id")

def get_description():
    """
    Retrieve the description metadata entry from the document
    """
    return sd["metadata"].get("core").get("description")

def get_edited():
    """
    Retrieve the last edited timestamp from the sadface document
    """
    return sd["metadata"].get("core").get("edited")

def get_edge(edge_id):
    """
    Retrieve the edge dict identified by the supplied edge ID

    Returns: An edge dict
    """
    for edge in sd["edges"]:
        if edge_id == edge["id"]:
            return edge

def get_global_metadata(namespace=None, key=None):
    """
    Retrieve the document's metadata. Results depend upon the specificity of the request
    which must include atom_id.

    If neither namespace nor key are provided then all metadata associated with the document is returned
    If a namespace is provided then only the specified namespace is returned
    If a namespace + a key is provided then the value associated with that key is returned
    Else nothing is returned

    Returns: A metadata dict
    """
    if namespace is None:
        return sd.get("metadata")
    else:
        if key is None:
            return sd.get("metadata").get(namespace)
        else:
            return sd.get("metadata").get(namespace).get(key)



def get_node(node_id):
    """
    Given a node's ID but no indication of node type, return the node if 
    it exists or else indicate that it doesn't to the caller.

    Returns: A node dict or None
    """
    for node in sd["nodes"]:
        if node_id == node["id"]:
            return node

def get_notes():
    """
    Retrieve the notes metadata entry from the document
    """
    return sd["metadata"].get("core").get("notes")

def get_resource(resource_id):
    """
    Retrieve the resource dict identified by the supplied resource ID

    Returns: An resource dict
    """
    for resource in sd["resources"]:
        if resource_id == resource["id"]:
            return resource

def get_resource_metadata(resource_id, namespace=None, key=None):
    """
    Retrieve a resource's metadata. Results depend upon the specificity of the request
    which must include a resource_id.

    If resource_id only is provided then all metadata assocaited with the resource is returned
    If reource_id + namespace is provided then only the specified namespace is returned
    If resource_id + namespace + a key is provided then the value associated with that key is returned
    Else nothing is returned

    Returns: A metadata dict
    """
    resource = get_resource(resource_id)
    if resource is not None:
        if namespace is None:
            return resource.get("metadata")
        else:
            if key is None:
                return resource.get("metadata").get(namespace)
            else:
                return resource.get("metadata").get(namespace).get(key)

def get_inference(inference_id):
    """
    Retrieve the inference dict identified by the supplied inference ID

    Returns: An inference dict if found otherwise None
    """
    for node in sd["nodes"]:
        if inference_id == node["id"]:
            return node

def get_inference_metadata(inference_id, namespace=None, key=None):
    """
    Retrieve an inference's metadata. Results depend upon the specificity of the request
    which must include a valid inference_id.

    If inference_id only is provided then all metadata associated with the scheme is returned
    If inference_id + namespace is provided then only the specified namespace is returned
    If inference_id + namespace + a key is provided then the value associated with that key is returned
    Else nothing is returned

    Returns: A metadata dict or None
    """
    inference = get_inference(inference_id)
    if inference is not None:
        if namespace is None:
            return inference.get("metadata")
        else:
            if key is None:
                return inference.get("metadata").get(namespace)
            else:
                return inference.get("metadata").get(namespace).get(key)

def get_source(atom_id, resource_id):
    """
    Retrieve the source dict identified by the supplied source ID

    Returns: An source dict
    """
    atom = get_atom(atom_id)
    if atom is not None:
        if len(atom["sources"]) > 0:
            for source in atom["sources"]:
                if resource_id == source["resource_id"]:
                    return source
            
def get_title():
    """
    Retrieve the title metadata entry from the document
    """
    return sd["metadata"].get("core").get("title")

def get_version():
    """
    Return a string indicating the version of the loaded SADFace document
    If no version parameter then version is less than <0.1
    Versioning introduced in v0.2 with introduction of version key to
        {"metadata":{ "core": { "version": "0.1"} }, ...}
    """
    version = sd.get("metadata").get("core").get("version")
    if None == version:
        print("no explicit version so 0.1")
        version = "0.1"

    return version


def import_json(json_string):
    """
    Take a string-encoded JSON document and loads it into a Python dict

    Returns: the loaded dict
    """
    return json.loads(json_string)

def initialise():
    """
    Reads the config file from the supplied location then uses the data
    contained therein to personalise a new SADFace document

    Returns: A Python dict representing the new SADFace document
    """
    global sd
    sd = new_sadface()
    return sd#new_sadface()

def list_atoms():
    """
    Return a list of atoms and their associated ID contained in the current 
    document, using the following format

    [ { 'id':'id-value', 'text':'text-value' } ]

    """
    atoms = []
    for node in sd["nodes"]:
        if "atom" == node["type"]:
            tmp = {}
            tmp["id"] = node["id"]
            tmp["text"] = node["text"]
            atoms.append(tmp)
    return atoms

def list_arguments():
    """
    Use to retrieve a list of arguments in the current document. Minimally
    we expect that each argument contains at least one sceheme node, so, by
    extension, a list of scheme nodes is also a list of arguments

    Returns: A list of scheme IDs
    """
    schemes = []
    for node in sd["nodes"]:
        if "scheme" == node["type"]:
            schemes.append(node.get("id"))
    return schemes

def list_conflict():
    """
    Use to retrieve a list of conflict nodes from the current document.

    Returns: A list of conflict IDs
    """
    conflicts = []
    for node in sd["nodes"]:
        if "conflict" == node["type"]:
            conflicts.append(node.get("id"))
    return conflicts

def list_resources():
    """
    Returns a comma separated list of resource IDs
    """
    return [ res["id"] for res in sd["resources"] ]

def list_inferences():
    """
    Use to retrieve a list of inference nodes from the current document.

    Returns: A list of inference IDs
    """
    inferences = []
    for node in sd["nodes"]:
        if "inference" == node["type"]:
            inferences.append(node.get("id"))
    return inferences


def load_from_file(filename):
    """
    Load the sadface document stored in the file identifed by the supplied
    filename.
    """
    try:
        with open(filename) as sadface_file:
            return json.load(sadface_file)
    except IOError:
        print("Could not read file:", filename)


def new_atom(text):
    """
    Creates a new SADFace atom node (Python dict) using the supplied text

    Returns: A Python dict representing the new SADFace atom
    """
    new_atom = {"id":new_uuid(), "type":"atom", "text":text, "sources":[], "metadata":{"core": {}}}
    return new_atom

def new_conflict(name):
    """
    Create a new SADFace conflict (Python dict) using the supplied conflict name. The conflict
    name should refer to a known conflict type from the set of {disagree | attack | defeat}

    Returns: A Python dict representing the new SADFace conflict
    """
    new_conflict = {"id":new_uuid(), "type":"conflict", "name":name, "metadata":{ "core": {}}}
    return new_conflict

def new_edge(source_id, target_id):
    """
    Creates & returns a new edge dict using the supplied source & 
    target IDs

    Returns: A Python dict representing the new edge
    """
    new_edge = {"id":new_uuid(), "source_id":source_id, "target_id":target_id}
    return new_edge

def new_sadface():
    """
    Creates & returns a new SADFace document

    Returns: A Python dict representing the new SADFace document
    """
    if config.location is None:
        new_doc = {"metadata":{ "core":{"version":version.__version__, "id":new_uuid(), "analyst_name":"A User", "analyst_email":"user@email.address", "created":now(), "edited":now()}}, "resources":[], "nodes":[], "edges":[]}

    else:
        new_doc = {"metadata":{ "core":{"version":version.__version__, "id":new_uuid(), "analyst_name":config.current.get("analyst", "name"), "analyst_email":config.current.get("analyst", "email"), "created":now(), "edited":now()}}, "resources":[], "nodes":[], "edges":[]}
    return new_doc

def new_resource(content):
    """
    Given the supplied content (Python String), create a new resource dict

    The arguments that SADFace describes are either constructed directly in a tool that writes
    them to this format, or else are sourced from elsewhere, e.g. an argumentative text or
    webpage, or else perhaps another medium, such as audio or video. Currently SADFace supports
    textual resources which are stored in the content key. Optionally a 
        "url":"some web location"
    pair can be added to the metadata to indicate a specific web location. Alternatively:
        "doi":"digital object identifier" - resolvable by dx.doi.org
        "magnet-link":"a torrent file"
        "isbn":"for books"
    Metadata may also store additional bibliographic or referencing/citation information
    as defined in bibtex formats.

    Returns: A Python dict representing the new SADFace resource
    """
    new_resource = {"id":new_uuid(), "content":content, "type":"text", "metadata":{ "core": {}}}
    return new_resource

def new_inference(name):
    """
    Create a new SADFace inference (Python dict) using the supplied scheme name. The scheme
    name should ideally refer to an existing scheme from a known schemeset

    Returns: A Python dict representing the instantiation of new SADFace inference
    """
    new_inference = {"id":new_uuid(), "type":"inference", "name":name, "metadata":{ "core": {}}}
    return new_inference

def new_source(resource_id, text, offset):
    """
    Create a new SADFace source (Python dict) using the supplied resource ID (a source always
    refers to an existing resource object) and identifying a section of text in the resource 
    as well as an offset for locating the text in the original resource.

    Text segment length is calculated from the length of the supplied text.

    As the resource object is enhanced to account for newer "types" of resource, so the
    source object must be enhanced to keep track and enable sources to index sub-parts of
    resources.

    Returns: A Python dict representing the new SADFace source
    """
    length = len(text)
    new_source = {"resource_id":resource_id, "text":text, "offset":offset, "length":length}
    return new_source

def new_uuid():
    """
    Utility method to generate a new universally unique ID. Used througout to uniquely
    identify various items such as atoms, schemes, resources, & edges

    Returns: A string
    """
    return str(uuid.uuid4())

def now():
    """
    Utility method to produce timestamps in ISO format without the microsecond
    portion, e.g. 2017-07-05T17:21:11

    Returns: A String
    """
    return datetime.datetime.now().replace(microsecond=0).isoformat()

def prettyprint(doc=None):
    """
    Retrieve a nicely formatted string encoded version of the SADFace document

    Returns: A String
    """
    string = sd
    if(doc is not None):
        string = doc
    return json.dumps(string, indent=4, sort_keys=True)

def print_doc(doc=None):
    """
    Retrieve a string encoded version of the SADFace document

    Returns: A String
    """
    string = sd
    if(doc is not None):
        string = doc
    return json.dumps(string,sort_keys=True)

def reset():
    """
    Return SADFace to it's initial state

    Basically delete, remove, or nullify any settings that have been applied.
    """
    global sd
    sd = {}
    config.reset()

def save(filename=None, filetype="json"):
    """
    Write the prettyprinted SADFace document to a file on disk.

    The default filetype is JSON. Other supported export formats are: 
    - Cytoscape
    - GraphViz/Dot
    
    """
    f = filename
    d = None
    pathname = None

    if filename is None:       
        if config.current is not None:
            f = config.current.get("file","name")
            d = config.current.get("file","dir")
            outname = f+d
        else:
            f = "out"
            d = ""
            pathname = f
    else:
        pathname = f


    if ("dot" == filetype):
        suffix = '.dot'
        if not pathname.endswith(suffix):
            pathname += suffix

        with codecs.open(pathname, 'w', 'utf-8') as outfile:
            outfile.write(export_dot())

    elif("cytoscape" == filetype):
        suffix = '.json'
        if not pathname.endswith(suffix):
            pathname += suffix
        
        with codecs.open(pathname, 'w', 'utf-8') as outfile:
            outfile.write(prettyprint(json.loads(export_cytoscape())))
    else:
        suffix = '.json'
        if not pathname.endswith(suffix):
            pathname += suffix
        
        with open(pathname, 'w') as outfile:
            json.dump(sd, outfile, indent=4, sort_keys=True, ensure_ascii=False)

def set_analyst(analyst):
    """
    sets the name of the argument analyst in the SADFace doc to the supplied name
    """
    sd["metadata"]["core"]["analyst_name"] = analyst
    
def set_atom_text(atom_id, new_text):
    """
    An atoms text key:value pair is the canonical representation of a portion of text 
    that exists in an argument. This should be updatable so that the overall document 
    makes sense. Links to original source texts are maintained via the source list 
    which indexes original text portions of linked resources.

    Returns: The updated atom dict
    """
    atom = get_atom(atom_id)
    if(atom is not None):
        atom["text"] = new_text
        return atom
    else:
        raise Exception("Could not set the text value for atom: "+atom_id)

def set_doc(newdoc):
    """
    Replaces the existing SADFace document with the supplied $newdoc
    """
    global sd
    sd = newdoc

def set_title(text):
    """
    Set a metadata entry for the document that contains a title. This is a
    useful but non-essential addendum to the base sadface document when
    building systems that make use of sadface.
    """
    sd["metadata"]["core"]["title"] = text

def set_claim(atom_id):
    """
    Enables a given atom to be nominated as the claim for the argument captured
    by this sadface document. A useful way to explicitly set the atom that should
    be considered to be the main claim.
    """
    atom = get_atom(atom_id)
    if(atom is not None): 
        sd["metadata"]["core"]["claim"] = atom_id
    else:
        raise Exception("Can't make atom ("+atom_id+") a claim because it doesn't exist")

def set_conflict_name(conflict_id, conflict_name):
    """
    Given an ID for an existing conflict node, set the name associated with it and return the 
    updated conflict node.
    
    Returns: Updated conflict dict
    """
    conflict = get_conflict(conflict_id)
    if(conflict is not None):
        conflict["name"] = conflict_name
        return conflict
    else:
        raise Exception("Could not set the name of conflict: "+conflict_id)

def set_created(timestamp):
    """
    Sets the creation timestamp for the SADFace document to the supplied timestamp.
    This can be useful when moving analysed argument data between formats whilst
    maintaining original metadata.
    """
    sd["metadata"]["core"]["created"] = timestamp

def set_description(text):
    """
    Set a metadata entry for the document that contains a description.
    """
    sd["metadata"]["core"]["description"] = text

def set_edited(timestamp):
    """
    Set the last edited timestamp for the SADFace doc to match the supplied
    timestamp. This can be useful when moving analysed argument data between formats 
    whilst maintaining original metadata.
    """
    sd["metadata"]["core"]["edited"] = timestamp

def set_document_id(id):
    """
    Set the SADFace document ID to match the supplied ID. This can be useful when 
    moving analysed argument data between formats whilst maintaining original metadata.
    """
    sd["metadata"]["core"]["id"] = id

def set_inference_name(inference_id, scheme_name):
    """
    Given an ID for an existing inference node, set the scheme name associated with it 
    and return the inference node.
    
    Returns: Updated inference dict
    """
    inference = get_inference(inference_id)
    if(inference is not None):
        inference["name"] = scheme_name
        return inference
    else:
        raise Exception("Could not set the name of inference: "+inference_id)

def update():
    """
    Updates the last edited timestamp for the SADFace doc to now
    """
    sd["metadata"]["core"]["edited"] = now()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open('README.md', "r", encoding="utf-8") as f:
    long_description = f.read()

exec(open('sadface/version.py').read())

setup(
    name='sadface',
    description='SADFace - The Simple Argument Description Format',
	long_description_content_type='text/markdown',
    long_description=long_description,
    author='Simon Wells',
    url='https://github.com/Open-Argumentation/SADFace.git',
    author_email='mail@simonwells.org',
	version=__version__,
    packages=find_packages(exclude=('deploy', 'etc', 'examples')),
)



import configparser
import json
import os
import unittest

import sadface as sf

class TestInitWithConfig(unittest.TestCase):
    def setUp(self):
        sf.reset()

    def tearDown(self):
        sf.reset()

    def test_config_load(self):
        """
        TESTS: sadface.config.load()
        """

        # No specified configuration file. location is None
        with self.assertRaises(Exception) as context:
            sf.config.load()
            
        self.assertTrue("Tried to load config file but location is set to None" in str(context.exception))

        # With bad configuration file
        sf.config.set_location("examples/minimal.py")
        with self.assertRaises(SystemExit):
            sf.config.load()
        
        # With good configuration file
        sf.config.set_location("etc/test.cfg")
        sf.config.load()
        sf.initialise()
        out = sf.get_document()
        out['metadata']['core']['created'] = ""
        out['metadata']['core']['edited'] = ""
        out['metadata']['core']['id'] = ""
        expected = json.loads('{ "edges": [], "metadata": { "core": { "analyst_email": "you-killed-my-father@prepare-to-die.com", "analyst_name": "Inigo Montoya", "created": "", "edited": "", "id": "", "version": "'+sf.version.__version__+'" } }, "nodes": [], "resources": []}')
        self.assertEqual(out, expected)


    def test_config_reset(self):
        """
        TESTS: sadface.config.reset()
        """
        current = None
        location = None

        sf.config.reset()
        self.assertEqual(current, sf.config.current)
        self.assertEqual(location, sf.config.location)

    def test_config_setlocation(self):
        """
        TESTS: sadface.config.set_location()
        """
        location = None
        self.assertEqual(location, sf.config.location)

        new_location = "src/sadface/sadface.py"
        sf.config.set_location(new_location)
        self.assertNotEqual(location, sf.config.location)
        self.assertEqual(new_location, sf.config.location)
        

    def test_init_with_config(self):
        """
        Tests: sadface.init() 
        
        Uses a user specified configuration that is different to the default
        """
        sf.config.set_location("etc/test.cfg")
        sf.config.load()
        sf.initialise()
        out = sf.get_document()
        out['metadata']['core']['created'] = ""
        out['metadata']['core']['edited'] = ""
        out['metadata']['core']['id'] = ""
        expected = json.loads('{ "edges": [], "metadata": { "core": { "analyst_email": "you-killed-my-father@prepare-to-die.com", "analyst_name": "Inigo Montoya", "created": "", "edited": "", "id": "", "version": "'+sf.version.__version__+'" } }, "nodes": [], "resources": []}')
        self.assertEqual(out, expected)

    def test_combined_config_init_method(self):
        """

        """
        sf.config.init("etc/test.cfg")
        sf.initialise()
        out = sf.get_document()
        out['metadata']['core']['created'] = ""
        out['metadata']['core']['edited'] = ""
        out['metadata']['core']['id'] = ""
        expected = json.loads('{ "edges": [], "metadata": { "core": { "analyst_email": "you-killed-my-father@prepare-to-die.com", "analyst_name": "Inigo Montoya", "created": "", "edited": "", "id": "", "version": "'+sf.version.__version__+'" } }, "nodes": [], "resources": []}')
        self.assertEqual(out, expected)


class TestInit(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass
    
    def test_init(self):
        """
        Tests: sadface.init()
        """
        sf.initialise()
        out = sf.get_document()
        out['metadata']['core']['created'] = ""
        out['metadata']['core']['edited'] = ""
        out['metadata']['core']['id'] = ""

        expected = json.loads('{ "edges": [], "metadata": { "core": { "analyst_email": "user@email.address", "analyst_name": "A User", "created":"", "edited":"", "id":"", "version": "'+sf.version.__version__+'" } }, "nodes": [], "resources": []}')
        self.assertEqual(out, expected)


class TestConfig(unittest.TestCase):
    def setUp(self):
        sf.reset()        

    def tearDown(self):
        sf.reset()
    
    def test_set_location(self):
        """
        Tests: sadface.config.set_location()

        """
        sf.config.set_location("etc/defaults.cfg")
        out = sf.config.location
        expected = "etc/defaults.cfg"
        self.assertEqual(out, expected)

class TestLoad(unittest.TestCase):
    def setUp(self):
        sf.reset()
        sf.config.set_location("etc/defaults.cfg")

    def tearDown(self):
        sf.reset()

    def test_load(self):
        """
        Tests: sadface.config.load()

        """
        sf.config.load()
        out = sf.config.current.get("analyst", "name")
        expected = 'A User'
        self.assertEqual(out, expected)

class TestLoadNoConfig(unittest.TestCase):
    def setUp(self):
        sf.reset()

    def tearDown(self):
        sf.reset()
    
    def test_load_no_config(self):
        """
        Tests: sadface.config.load()

        Ensure that if load is called but location isn't set then an
        exception must be raised
        """
        with self.assertRaises(Exception) as context:
            sf.config.load()
        out = str(context.exception)
        expected = 'Tried to load config file but location is set to None'
        self.assertEqual(out, expected)


if __name__ == "__main__":
    
    unittest.main()
